﻿#ifndef _MATE1PLY_H_
#define _MATE1PLY_H_

#include "../shogi.h"
#ifdef MATE_1PLY

// 超高速1手詰め判定ライブラリ
// cf. 新規節点で固定深さの探索を併用するdf-pnアルゴリズム gpw05.pdf


// 方角を表す。遠方駒の利きや、玉から見た方角を表すのに用いる。
// bit0..右上、bit1..右、bit2..右下、bit3..上、bit4..下、bit5..左上、bit6..左、bit7..左下
enum Directions : uint8_t {
  DIRECTIONS_RU = 1 , DIRECTIONS_R  = 2  , DIRECTIONS_RD = 4  , DIRECTIONS_U   = 8 ,
  DIRECTIONS_D  =16 , DIRECTIONS_LU = 32 , DIRECTIONS_L  = 64 , DIRECTIONS_LD = 128 };
const int DIRECTIONS_NB = 256; // ↑のenumで表現できない範囲なので出しておく。

// 方角 (Directionsをpopしたもの)
enum Direct { DIRECT_NB = 8 };
inline bool is_ok(Direct d) { return 0 <= d && d < DIRECT_NB; }
inline Direct pop_directions(uint32_t& d) { return (Direct)pop_lsb(d); }

// DirectをSquare型の差分値で表現したもの
const Square DirectToDelta_[DIRECT_NB] = { DELTA_SE,DELTA_E,DELTA_NE,DELTA_S,DELTA_N,DELTA_SW,DELTA_W,DELTA_NW, };
inline Square DirectToDelta(Direct d) { ASSERT_LV3(is_ok(d));  return DirectToDelta_[d]; }

// 1手詰め判定高速化テーブル
struct alignas(8) MateInfo
{
  union {
    // この形において詰ませるのに必要な駒種
    uint8_t/*HandKind*/ hand_kind;

    // 玉から見てどの方向に打てば詰むのかの情報(要素[0]がもったいないので、unionにしてこの要素はhand_kindと共用にする。
    Directions directions[PIECE_RAW_NB];
  };
};


#ifdef USE_AVX2
using ymm = __m256i;
static const ymm ymm_one = _mm256_set1_epi8(1);
static const ymm ymm_zero = _mm256_setzero_si256();
#endif

// 利きの数や遠方駒の利きを表現するByteBoard
// 玉の8近傍を回収するなど、アライメントの合っていないアクセスをするのでこの構造体にはalignasをつけないことにする。
struct ByteBoard
{
  // ある升の利きの数
  uint8_t count(Square sq) const { return e[sq]; }

  // ある升にある長い利きの方向
  Directions directions(Square sq) const { return (Directions)e[sq]; }

  // ゼロクリア
  void clear() { memset(e, 0, 128); }

  // ある升の周辺8近傍の利きを取得。1以上の値のところが1になる。壁のところは0になる。
  uint8_t around8(Square sq) const {
#ifdef USE_AVX2
    // メモリアクセス違反であるが、StateInfoのなかで使うのでお隣のメモリが
    // 10bytesぐらい確保されているだろうから問題ない。
    ymm y0 = _mm256_load_si256((__m256i*)&e[sq - 10]);
    ymm mask0 = _mm256_cmpgt_epi8(y0, ymm_zero);
    uint32_t m0 = _mm256_movemask_epi8(mask0); // sqの升の右上の升から32升分は取得できたので、これをPEXTで回収する。
    return PEXT32(m0, 0b111000000101000000111);
#else
    uint8_t a8 = 0;
    if (e[sq- 10]) a8 |=   1;
    if (e[sq-  9]) a8 |=   2;
    if (e[sq-  8]) a8 |=   4;
    if (e[sq-  1]) a8 |=   8;
    if (e[sq+  1]) a8 |=  16;
    if (e[sq+  8]) a8 |=  32;
    if (e[sq+  9]) a8 |=  64;
    if (e[sq+ 10]) a8 |= 128;
    return (uint8_t)a8;
#endif
  }

  // ある升の周辺8近傍の利きを取得。2以上の値のところが1になる。壁のところは0になる。
  uint8_t around8_larger_than_two(Square sq)
  {
#ifdef USE_AVX2
    ymm y0 = _mm256_load_si256((__m256i*)&e[sq - 10]);
    ymm mask0 = _mm256_cmpgt_epi8(y0, ymm_one);
    uint32_t m0 = _mm256_movemask_epi8(mask0); // sqの升の右上の升から32升分は取得できたので、これをPEXTで回収する。
    return PEXT32(m0, 0b111000000101000000111);
#else
    uint8_t a8 = 0;
    if (e[sq - 10] > 1 ) a8 |=   1;
    if (e[sq - 9]  > 1 ) a8 |=   2;
    if (e[sq - 8]  > 1 ) a8 |=   4;
    if (e[sq - 1]  > 1 ) a8 |=   8;
    if (e[sq + 1]  > 1 ) a8 |=  16;
    if (e[sq + 8]  > 1 ) a8 |=  32;
    if (e[sq + 9]  > 1 ) a8 |=  64;
    if (e[sq + 10] > 1 ) a8 |= 128;
    return (uint8_t)a8;
#endif
  }

  // 各升の利きの数 or Directions
  uint8_t e[SQ_NB];

  // around8()などである升の8近傍の情報を回収したときに壁の位置のマスクが欲しいときがあるから、そのマスク情報。
  static uint8_t NEIGHBOR_MASK[SQ_NB];
};

#endif // MATE_1PLY
#endif // _MATE1PLY_H_
